"""Release [0.50.0].

Revision ID: 0.50.0
Revises: 389046140cad
Create Date: 2023-11-27 18:29:12.329006

"""


# revision identifiers, used by Alembic.
revision = "0.50.0"
down_revision = "389046140cad"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
