"""Release [0.47.0].

Revision ID: 0.47.0
Revises: 722392c91006
Create Date: 2023-11-14 16:10:13.251399

"""


# revision identifiers, used by Alembic.
revision = "0.47.0"
down_revision = "722392c91006"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
