"""Release [0.52.0].

Revision ID: 0.52.0
Revises: e5225281b4d3
Create Date: 2023-12-12 17:15:11.532341

"""


# revision identifiers, used by Alembic.
revision = "0.52.0"
down_revision = "e5225281b4d3"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
