"""add connector skew tolerance [e5225281b4d3].

Revision ID: e5225281b4d3
Revises: 0.51.0
Create Date: 2023-12-08 13:50:25.321142

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "e5225281b4d3"
down_revision = "0.51.0"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("service_connector", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column("expires_skew_tolerance", sa.Integer(), nullable=True)
        )

    # ### end Alembic commands ###


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("service_connector", schema=None) as batch_op:
        batch_op.drop_column("expires_skew_tolerance")

    # ### end Alembic commands ###
