#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Initialization of Kserve Secret Schemas.

These are secret schemas that can be used to authenticate Kserve to the
Artifact Store used to store served ML models.
"""
from zenml.integrations.kserve.secret_schemas.secret_schemas import (
    KSERVE_AZUREBLOB_SECRET_SCHEMA_TYPE,
    KSERVE_GS_SECRET_SCHEMA_TYPE,
    KSERVE_S3_SECRET_SCHEMA_TYPE,
    KServeAzureSecretSchema,
    KServeGSSecretSchema,
    KServeS3SecretSchema,
)

__all__ = [
    "KSERVE_AZUREBLOB_SECRET_SCHEMA_TYPE",
    "KSERVE_GS_SECRET_SCHEMA_TYPE",
    "KSERVE_S3_SECRET_SCHEMA_TYPE",
    "KServeAzureSecretSchema",
    "KServeGSSecretSchema",
    "KServeS3SecretSchema",
]
