"""Release 0.45.1 [0.45.1].

Revision ID: 0.45.1
Revises: 6f707b385dc1
Create Date: 2023-10-15 14:37:18.917843

"""


# revision identifiers, used by Alembic.
revision = "0.45.1"
down_revision = "6f707b385dc1"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
