"""Release [0.45.4].

Revision ID: 0.45.4
Revises: 4f66af55fbb9
Create Date: 2023-10-20 12:27:07.315167

"""


# revision identifiers, used by Alembic.
revision = "0.45.4"
down_revision = "4f66af55fbb9"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
